//------------------------------------------------------------------------------
//
//  npctrl.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2011 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------

#include "npdata.h"
#include "npio.h"
#include "npctrl.h"

#include "data/npnode.h"
#include "io/npgl.h"
#include "io/npfile.h"
#include "io/npmouse.h"
#include "ctrl/npengine.h"

#include "windows.h"

void npCtrlFile (int command, void* dataRef);

void npCtrlVector (int command, void* dataRef);
void npCtrlSelect (int command, void* dataRef);
void npCtrlProperty (int command, void* dataRef);
void npCtrlGlobal (int command, void* dataRef);

void TraverseTree (int command, NPnodePtr node, void* dataRef);

//------------------------------------------------------------------------------
void npInitCtrl (void* dataRef)
{
	npInitEngine (dataRef);
}

//------------------------------------------------------------------------------
void npCloseCtrl (void* dataRef)
{
	npCloseEngine (dataRef);
}

//------------------------------------------------------------------------------
int npAppLoop (void* dataRef)
{
	int err = 0;

#ifdef NP_LINUX_
//	MainAppLoopLinux();
#endif

#ifdef NP_MSW_
//	MainAppLoopMSW();
	npAppLoopGlut (dataRef);
#endif

#ifdef NP_OSX_
	//	MainAppLoopOSX();
#endif


	return err;
}



//------------------------------------------------------------------------------
void TraverseTree (int command, NPnodePtr node, void* dataRef)
{
	int i = 0;
	
	pData data = (pData) dataRef;

	// recursively iterate through all nodes in the tree
	for (i=0; i < node->childCount ; i++)
		TraverseTree (command, node->child[i], dataRef);
	
	if ( node->selected 
		 || command == kNPgraphSelectOn 
		 || command == kNPgraphSelectOff )
	{
		data->map.activeNode = node;
		npCtrlCommandSet (command, node, dataRef);
	}
}


//------------------------------------------------------------------------------
void npCtrlCommand (int command, void* dataRef)
{
	int i = 0;
	int indexTemp = 0;

	pData data = (pData) dataRef;

	NPnodePtr node = data->map.activeNode;
	NPnodePtr nodeParent = NULL;


	switch (command)
	{
		// kCmdGlobal... global commands, draw indicators/HUD, quit
//		case kQuit : npCtrlGlobal (command, dataRef); break;

		case kNPfileNew : npCtrlFile (command, dataRef); break;
		case kNPfileOpen : npCtrlFile (command, dataRef); break;
		case kNPfileClose : npCtrlFile (command, dataRef); break;
		case kNPfileSave : npCtrlFile (command, dataRef); break;
		case kNPfileSaveAs : npCtrlFile (command, dataRef); break;
		case kNPfileImport : npCtrlFile (command, dataRef); break;
		case kNPfileExport : npCtrlFile (command, dataRef); break;
			
		case kNPopenURL : npCtrlFile (command, dataRef); break;
		case kNPopenApp : npCtrlFile (command, dataRef); break;
		case kNPopenAntz : npCtrlFile (command, dataRef); break;

		case kNPgraphMenu : npCtrlGlobal (command, dataRef); break;
		case kFullScreenExit : npCtrlGlobal (command, dataRef); break;
		case kNPgraphClearFrameBuffer : npCtrlGlobal (command, dataRef); break;
		case kNPgraphBackground : npCtrlGlobal (command, dataRef); break;		
		
		case kNPgraphNormalMode : npCtrlGlobal (command, dataRef); break;
		case kNPgraphShaderMode : npCtrlGlobal (command, dataRef); break;
		case kNPgraphAlphaMode : npCtrlGlobal (command, dataRef); break;

		// kCmdSelect.. modify selection / scope (navigate the map/space)	
		// Selects the active axes and nodes to manipulate
		case kNPgraphSelectAxes : npCtrlSelect (command, dataRef); break;

		case kNPgraphGrid : npCtrlSelect (command, dataRef); break;
		case kNPgraphCamera : npCtrlSelect (command, dataRef); break;

		case kNPgraphNext : npCtrlSelect (command, dataRef); break;
		case kNPgraphNextBranch : npCtrlSelect (command, dataRef); break;

		case kNPgraphNew : npCtrlSelect (command, dataRef); break;
		case kNPgraphDelete : npCtrlSelect (command, dataRef); break;

		case kNPgraphSelect : npCtrlSelect (command, dataRef); break;	// broken, debug zz
		case kNPgraphSelectOn : npCtrlSelect (command, dataRef); break;	// broken, debug zz
		case kNPgraphSelectOff : npCtrlSelect (command, dataRef); break;	// broken, debug zz

		case kNPgraphSelectAll : npCtrlSelect (command, dataRef); break;		// broken, debug zz
		case kNPgraphSelectAllOff : npCtrlSelect (command, dataRef); break;
		case kNPgraphSelectNone : npCtrlSelect (command, dataRef); break;
		case kNPgraphSelectOne : npCtrlSelect (command, dataRef); break;
		case kNPgraphSelectTwo : npCtrlSelect (command, dataRef); break;
		case kNPgraphSelectThree : npCtrlSelect (command, dataRef); break;

		case kNPgraphPresetOne : npDataPreset (command, dataRef); break;
		case kNPgraphPresetTwo : npDataPreset (command, dataRef); break;
		case kNPgraphPresetThree : npDataPreset (command, dataRef); break;
		case kNPgraphPresetFour : npDataPreset (command, dataRef); break;
		case kNPgraphPresetFive : npDataPreset (command, dataRef); break;
		case kNPgraphPresetSix : npDataPreset (command, dataRef); break;
		case kNPgraphPresetSeven : npDataPreset (command, dataRef); break;
		case kNPgraphPresetEight : npDataPreset (command, dataRef); break;
		

		default : // not a global command

			// exclude global commands or only 1 selected node from applying to all nodes-
			// ie. allows for adding an unsellected node to the current group
			if ( !node->selected || data->map.nodeRootIndex < kNPnodeRootPin )
			{
				node = data->map.node[data->map.nodeRootIndex];
				npCtrlCommandSet (command, node, data);
			}
			else
			{											// apply to selected nodes
				indexTemp = data->map.nodeRootIndex;
	
				if (node->branchLevel > 0)
					nodeParent = node->parent;
	
				for (i=0; i < data->map.nodeRootCount; i++)	//search and apply to selected nodes
				{
					data->map.nodeRootIndex = i;
					TraverseTree (command, data->map.node[i], dataRef);
				}
					data->map.nodeRootIndex = indexTemp;
					data->map.activeNode = node;
			}	
		break;
	}
}


//------------------------------------------------------------------------------
void npCtrlCommandSet (int command, void* nodeRef, void* dataRef)
{
	pData data = (pData) dataRef;

	NPnodePtr node = (NPnodePtr) nodeRef;


	switch (command)
	{
		case kNPgraphSelect : npCtrlSelect (command, dataRef); break;	// broken, debug zz
		case kNPgraphSelectOn : npCtrlSelect (command, dataRef); break;	// broken, debug zz
		case kNPgraphSelectOff : npCtrlSelect (command, dataRef); break;	// broken, debug zz

//		case kNPgraphSelectAll : npCtrlSelect (command, dataRef); break;		// broken, debug zz

		//translate the selected objects relative to the camera
		// kCmd... apply to current selection / scope (X/Y/XY/Z/XYZ)
		case kNPgraphXdecrease : npCtrlVector (command, dataRef); break;
		case kNPgraphXdecreaseOff : npCtrlVector (command, dataRef); break;
		case kNPgraphXincrease : npCtrlVector (command, dataRef); break;
		case kNPgraphXincreaseOff : npCtrlVector (command, dataRef); break;

		case kNPgraphYdecrease : npCtrlVector (command, dataRef); break;
		case kNPgraphYdecreaseOff : npCtrlVector (command, dataRef); break;
		case kNPgraphYincrease : npCtrlVector (command, dataRef); break;
		case kNPgraphYincreaseOff : npCtrlVector (command, dataRef); break;

		case kNPgraphZdecrease : npCtrlVector (command, dataRef); break;
		case kNPgraphZdecreaseOff : npCtrlVector (command, dataRef); break;
		case kNPgraphZincrease : npCtrlVector (command, dataRef); break;
		case kNPgraphZincreaseOff : npCtrlVector (command, dataRef); break;

		case kNPgraphRotateLeft : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateLeftOff : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateRight : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateRightOff : npCtrlVector (command, dataRef); break;

		case kNPgraphRotateUp : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateUpOff : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateDown : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateDownOff : npCtrlVector (command, dataRef); break;

		case kNPgraphRotateCCW : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateCCWOff : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateCW : npCtrlVector (command, dataRef); break;
		case kNPgraphRotateCWOff : npCtrlVector (command, dataRef); break;

		case kNPgraphZoomOn : npCtrlVector (command, dataRef); break;
		case kNPgraphZoomOff : npCtrlVector (command, dataRef); break;

		// Change node property
		case kNPgraphChannelUp : npCtrlProperty (command, dataRef); break;
		case kNPgraphChannelDown : npCtrlProperty (command, dataRef); break;

		case kNPgraphColorUp : npCtrlProperty (command, dataRef); break;
		case kNPgraphColorDown : npCtrlProperty (command, dataRef); break;
		case kNPgraphAlphaUp : npCtrlProperty (command, dataRef); break;
		case kNPgraphAlphaDown : npCtrlProperty (command, dataRef); break;
			
		case kNPgraphTexture : npCtrlProperty (command, dataRef); break;
		case kNPgraphCenter : npCtrlProperty (command, dataRef); break;
		case kNPgraphLines : npCtrlProperty (command, dataRef); break;
		case kNPgraphPoints : npCtrlProperty (command, dataRef); break;
		case kNPgraphShader : npCtrlProperty (command, dataRef); break;
		case kNPgraphPrimitive : npCtrlProperty (command, dataRef); break;
		case kNPgraphTopo : npCtrlProperty (command, dataRef); break;
		case kNPgraphMesh : npCtrlProperty (command, dataRef); break;
		case kNPgraphScroll : npCtrlProperty (command, dataRef); break;
		case kNPgraphFreeze : npCtrlProperty (command, dataRef); break;
		case kNPgraphHide : npCtrlProperty (command, dataRef); break;
		case kNPgraphClear : npCtrlProperty (command, dataRef); break;

		case kNPgraphDrawIndicators : npCtrlProperty (command, dataRef); break;

		case kNPgraphSegments : npCtrlProperty (command, dataRef); break;

		case kEventLowSetpoint : npCtrlProperty (command, dataRef); break;
		case kEventHighSetpoint : npCtrlProperty (command, dataRef); break;

		case kNPgraphRatio : npCtrlProperty (command, dataRef); break;	
			
		default : break;
	}
}


//------------------------------------------------------------------------------
void npCtrlFile (int command, void* dataRef)
{
	pData data = (pData) dataRef;

	switch (command)
	{
		case kNPfileNew : npFileDialog (NULL, kNPfileDialogNew, dataRef); break;

		case kNPfileOpen : npFileOpenMap ("ANTZ0001.CSV", 1, 12, dataRef); break; //npFileDialog (NULL, kNPfileDialogOpen, dataRef); break;
		case kNPfileClose : npFileDialog (NULL, kNPfileDialogClose, dataRef); break;

		case kNPfileSave : npFileSaveMap  ("ANTZ0001.CSV", 1, 12, dataRef); break; //npFileDialog (NULL, kNPfileDialogSave, dataRef); break;
		case kNPfileSaveAs : npFileDialog (NULL, kNPfileDialogSaveAs, dataRef); break;

		case kNPfileMapOne :
			if (data->io.key.modShift)
				npFileSaveMap  ("ANTZ0001.CSV", 1, 12, dataRef);
			else
				npFileOpenMap ("ANTZ0001.CSV", 1, 12, dataRef);
			break;
		case kNPfileMapTwo :
			if (data->io.key.modShift)
				npFileSaveMap  ("ANTZ0002.CSV", 1, 12, dataRef);
			else
				npFileOpenMap ("ANTZ0002.CSV", 1, 12, dataRef);
			break;
		case kNPfileMapThree :
			if (data->io.key.modShift)
				npFileSaveMap  ("ANTZ0003.CSV", 1, 12, dataRef);
			else
				npFileOpenMap ("ANTZ0003.CSV", 1, 12, dataRef);			//zz end
			break;

		case kNPfileImport : npSetChannelFile ("TNG00001.CSV", 1, 12, dataRef); break; //npFileDialog (NULL, kNPfileDialogImport, dataRef); break;
		case kNPfileExport : npFileDialog (NULL, kNPfileDialogExport, dataRef); break;

		case kNPopenURL : npOpenURL (NULL, dataRef); break;
		case kNPopenApp : npOpenApp (NULL, dataRef); break;
		case kNPopenAntz : npOpenAntz (NULL, dataRef); break;

		default : break;
	}
}


// update the translateVect for selected objects
// translateVect is relative to global coordinates
// translateRate is relative to the camera
//------------------------------------------------------------------------------
void npCtrlVector (int command, void* dataRef)
{
	float velocity = 0.0f;

	NPnodePtr node = NULL;

	pData data = (pData) dataRef;

	// apply command to activeNode
	node = data->map.activeNode;

	// if shift pressed then use fast velocity, otherwise slow
	if (data->io.key.modShift)
		velocity = data->ctrl.fast;
	else
		velocity = data->ctrl.slow;

	switch (command)
	{								//normalize and change to 0.0f - velocity,	debug zz
		case kNPgraphXdecrease : node->translateRate.x = -0.25f * velocity; break;
		case kNPgraphXdecreaseOff : 
			if (node->translateRate.x < 0.0f)	//graceful multi-key behavior
				node->translateRate.x = 0.0f; 
			break;
		case kNPgraphXincrease : node->translateRate.x = 0.25f * velocity; break;
		case kNPgraphXincreaseOff : 
			if (node->translateRate.x > 0.0f)
				node->translateRate.x = 0.0f; 
			break;

		case kNPgraphYdecrease : node->translateRate.y = -0.25f * velocity; break;
		case kNPgraphYdecreaseOff : 
			if (node->translateRate.y < 0.0f)
				node->translateRate.y = 0.0f;
			break;
		case kNPgraphYincrease : node->translateRate.y = 0.25f * velocity; break;
		case kNPgraphYincreaseOff : 
			if (node->translateRate.y > 0.0f)
				node->translateRate.y = 0.0f;
			break;

		case kNPgraphZdecrease : node->translateRate.z = -0.25f * velocity; break;
		case kNPgraphZdecreaseOff : 
			if (node->translateRate.z < 0.0f)
				node->translateRate.z = 0.0f;
			break;
		case kNPgraphZincrease : node->translateRate.z = 0.25f * velocity; break;
		case kNPgraphZincreaseOff : 
			if (node->translateRate.z > 0.0f)
				node->translateRate.z = 0.0f;
			break;
		
		case kNPgraphRotateLeft : node->rotateRate.y = -0.25f * velocity; break;
		case kNPgraphRotateLeftOff : 
			if (node->rotateRate.y < 0.0f)
				node->rotateRate.y = 0.0f;
			break;
		case kNPgraphRotateRight : node->rotateRate.y = 0.25f * velocity; break;
		case kNPgraphRotateRightOff : 
			if (node->rotateRate.y > 0.0f)
				node->rotateRate.y = 0.0f;
			break;

		case kNPgraphRotateUp : node->rotateRate.x = 0.25f * velocity; break;
		case kNPgraphRotateUpOff :
			if (node->rotateRate.x > 0.0f)
				node->rotateRate.x = 0.0f;
			break;
		case kNPgraphRotateDown : node->rotateRate.x = -0.25f * velocity; break;
		case kNPgraphRotateDownOff :
			if (node->rotateRate.x < 0.0f)
				node->rotateRate.x = 0.0f;
			break;

		case kNPgraphRotateCCW : node->rotateRate.z = -0.25f * velocity; break;
		case kNPgraphRotateCCWOff : 
			if (node->rotateRate.z < 0.0f)
				node->rotateRate.z = 0.0f;
			break;
		case kNPgraphRotateCW : node->rotateRate.z = 0.25f * velocity; break;
		case kNPgraphRotateCWOff : 
			if (node->rotateRate.z > 0.0f)
				node->rotateRate.z = 0.0f;
			break;

		case kNPgraphZoomOn :
			if (data->io.key.modShift)		//scales the object
			{
				if (data->io.axes.x == true )
					node->scaleRate.x = -0.003f;
				if (data->io.axes.y == true )
					node->scaleRate.y = -0.003f;
				if (data->io.axes.z == true )
					node->scaleRate.z = -0.003f;			
			}
			else
			{
				if (data->io.axes.x == true )
					node->scaleRate.x = 0.003f;
				if (data->io.axes.y == true )
					node->scaleRate.y = 0.003f;
				if (data->io.axes.z == true )
					node->scaleRate.z = 0.003f;			
				}
			break;
		case kNPgraphZoomOff :
			node->scaleRate.x = 0.0f;
			node->scaleRate.y = 0.0f; 
			node->scaleRate.z = 0.0f; 
			break;

		default : break;
	}
}

//------------------------------------------------------------------------------
void npCtrlSelect (int command, void* dataRef)
{
	int i = 0;
	int commandTemp = 0;

	NPnodePtr node = NULL;
	NPnodePtr nodeChild = NULL;
	NPnodePtr nodeParent = NULL;

	pData data = (pData) dataRef;

	node = data->map.activeNode;


	switch (command)
	{						// add ability to copy existing node parameters, debug zz
		case kNPgraphNew :
			//if camera selected and no pins exist then create 1st pin
			if (data->map.nodeRootIndex == kNPnodeRootCamera
				&& data->map.nodeRootCount <= kNPnodeRootPin )
			{
				node = npNodeNew (kNodePin, NULL, dataRef);
				npNodeNew (node->type, node, dataRef);
			}
			else
			{
				//if camera or grid
				if (node->type == kNodeCamera || node->type == kNodeGrid)
				{
					if ( node->branchLevel == 1  && !data->io.key.modShift )
						node = npNodeNew (node->type, node->parent, dataRef);
					else			
						node = npNodeNew (node->type, node, dataRef);
				}
				else if (node->type == kNodePin)
				{	//if root create new pin, if shift create child
					if (node->branchLevel == 0 && !data->io.key.modShift
						&& node->childCount > 0 )
					{
						node = npNodeNew (kNodePin, NULL, dataRef);
						npNodeNew (node->type, node, dataRef);
					} else
						node = npNodeNew (node->type, node, dataRef);
				}
			}
			npSetCamTarget();	//sets camera target to selected node

			if (node->type == kNodeCamera)
				printf("New camera ");
			if (node->type == kNodeGrid)
				printf("New grid ");
			if (node->type == kNodePin)
				printf("New pin ");
			printf("id: %d   branchLevel: %d\n", node->id, node->branchLevel);
			break;

		case kNPgraphDelete :
			npNodeDelete (node, dataRef);
			npSetCamTarget();	//sets camera target to selected node
			break;
			
		case kNPgraphNext :
			if (data->map.nodeRootCount <= kNPnodeRootPin)		// if no nodes select cam
			{
				node = data->map.activeCam;
				data->map.activeNode = node;
				data->map.nodeRootIndex = kNPnodeRootCamera;
				break;
			}
			if (data->map.nodeRootIndex < kNPnodeRootPin)		// camera or grid selected
			{
				// restore previously selected node
				data->map.nodeRootIndex = data->map.selectedIndex;
				data->map.activeNode = data->map.selectedPinNode;
			}
			else
			{
				if (node->branchLevel < 1)				// root level pin
				{
					// change our root node selection
					if  (data->io.key.modShift)
						data->map.nodeRootIndex--;
					else
						data->map.nodeRootIndex++;

					// keep inbounds and don't select camera or grid
					if (data->map.nodeRootIndex >= data->map.nodeRootCount)
							data->map.nodeRootIndex = kNPnodeRootPin;
					else if (data->map.nodeRootIndex < kNPnodeRootPin)	// select last pin
						data->map.nodeRootIndex = data->map.nodeRootCount - 1;

					// select our next root node and make active
					data->map.selectedIndex = data->map.nodeRootIndex;
					node = data->map.node[data->map.selectedIndex];
					data->map.activeNode = node;
					data->map.selectedPinNode = data->map.activeNode;

					// set the examiner target on the selected root node
					npSetCamTarget();
				}
				else
				{						// if available, select next sibling
					nodeChild = node;
					node = nodeChild->parent;

					if  (data->io.key.modShift)
						node->childIndex--;
					else
						node->childIndex++;
		
					// keep in bounds and roll over between first and last
					if (node->childIndex >= node->childCount)
						node->childIndex = 0;
					else if (node->childIndex < 0)
						node->childIndex = node->childCount - 1;

					// select our next sibling node and make active
					node = node->child[node->childIndex];
					data->map.activeNode = node;
					data->map.selectedPinNode = data->map.activeNode;
				}
			}
			node = data->map.activeNode;

			if (node->recordID)
				printf ("pin id: %d   recordID: %d\n", node->id, node->recordID);
			else
				printf ("pin id: %d\n", node->id);
			break;

		case kNPgraphNextBranch : 
			if (data->io.key.modShift)
			{
				if (node->branchLevel > 0)		// do nothing if at the trunk
				{
					// traverse up 1 level to the childs parent
					if (node->parent == NULL)
						printf ("err 3582 - child node has NULL parent\n");
					else
						data->map.activeNode = node = node->parent;
				}
			}		// traverse down 1 level if selected child exists
			else if (node->child[node->childIndex] != NULL)
			{
				node = node->child[node->childIndex];
				data->map.activeNode = node;
			}

			if (data->map.nodeRootIndex >= kNPnodeRootPin)
				data->map.selectedPinNode = data->map.activeNode;
			else if (data->map.nodeRootIndex == kNPnodeRootCamera)
				data->map.activeCam = node;
			else if (data->map.nodeRootIndex == kNPnodeRootGrid)
				data->map.selectedGrid = node;

			
			if (node->type == kNodeCamera)
				printf("camera ");
			if (node->type == kNodeGrid)
				printf("grid ");
			if (node->type == kNodePin)
				printf("pin ");

			if (node->recordID)
				printf ("id: %d   branchLevel: %d   recordID: %d\n",
						node->id, node->branchLevel, node->recordID);
			else
				printf ("id: %d   branchLevel: %d\n",
						node->id, node->branchLevel);
			break;
				
		case kNPgraphGrid :
			if (data->map.nodeRootIndex == kNPnodeRootGrid)	// grid selected
			{
				if (node == data->map.node[kNPnodeRootGrid])
				{
					if (node->child[0] != NULL)
					{
						node->childIndex = 0;		//select first child cam
						node = node->child[node->childIndex];
					}
				}
				else 
				{
					nodeParent = node->parent;
					if (nodeParent != NULL)
					{								// select next grid
						nodeParent->childIndex++;
						if (nodeParent->childIndex >= nodeParent->childCount)
						{
							nodeParent->childIndex = 0;
							node = data->map.node[kNPnodeRootGrid];
						}
						else
							node = nodeParent->child[nodeParent->childIndex];
					}
					else	//only happens if addtiional root cameras created
					{
						node = data->map.node[kNPnodeRootGrid];
						printf ("Selecting Root Grid id: %d\n", node->id);
					}
				}
				data->map.selectedGrid = node;		//update grid selection
				data->map.activeNode = node;
			}
			else	// make the selected grid active
			{
				node = data->map.selectedGrid;
				if (node == NULL)
					node = data->map.node[kNPnodeRootGrid];

				data->map.selectedGrid = node;
				data->map.activeNode = node;
				data->map.nodeRootIndex = kNPnodeRootGrid;
			}

			printf ("grid id: %d\n", node->id);
			break;

		case kNPgraphCamera :
			if (data->map.nodeRootIndex == kNPnodeRootCamera)	// camera selected
			{
				nodeParent = node->parent;
				// reset the camera if shift held, else switch cameras
				if (data->io.key.modShift)
				{
					if (node->parent != NULL)
						npDataCameraPreset (nodeParent->childIndex, node, data);
					else
						npDataCameraPreset (0, node, data);

					printf ("Reset ");
				}
				else
				{
					if (node == data->map.node[kNPnodeRootCamera])
					{
						if (node->child[0] != NULL)
						{
							node->childIndex = 0;		//select first child cam
							node = node->child[node->childIndex];
						}
					}
					else 
					{
						if (nodeParent != NULL)
						{	// select next camera
							nodeParent->childIndex++;
							if (nodeParent->childIndex >= nodeParent->childCount)
							{
								nodeParent->childIndex = 0;
								node = data->map.node[kNPnodeRootCamera];
							}
							else
								node = nodeParent->child[nodeParent->childIndex];
						}
						else	//only happens if addtiional root cameras created
						{
							node = data->map.node[kNPnodeRootCamera];
							printf ("Aux Root Camera Selected id: %d\n", node->id);
						}
					}
					// set active camera and node
					data->map.activeCam = node;
					data->map.activeNode = data->map.activeCam;
				}
			}
			else	// select the current camera if already selected
			{
				node = data->map.activeCam;
				if (node == NULL)
					node = data->map.node[kNPnodeRootCamera];
				
				//reset the camera if shift held, else switch cameras
				nodeParent = node->parent;
				if (data->io.key.modShift)
				{
					if (nodeParent != NULL)
						npDataCameraPreset (nodeParent->childIndex, node, data);
					else
						npDataCameraPreset (0, node, data);
				}
				else
				{
					data->map.activeCam = node;
					data->map.activeNode = node;
					data->map.nodeRootIndex = kNPnodeRootCamera;
				}
			}
			printf ("camera id: %d\n", node->id);
			break;

		case kNPgraphSelectAxes :
			if (data->io.axes.x == true 
				&& data->io.axes.y == true
				&& data->io.axes.z == true )
			{
				data->io.axes.z = false;
				printf ("axes: XY\n");				// make part of HUD debug, zz
			}
			else if (data->io.axes.x == true 
					 && data->io.axes.y == true
					 && data->io.axes.z == false )
			{
				data->io.axes.x = true; 
				data->io.axes.y = false; 
				data->io.axes.z = false;
				printf ("axes: X\n");
			}
			else if (data->io.axes.x == true 
					 && data->io.axes.y == false
					 && data->io.axes.z == false )
			{
				data->io.axes.x = false; 
				data->io.axes.y = true; 
				data->io.axes.z = false;
				printf ("axes: Y\n");
			}
			else if (data->io.axes.x == false 
					 && data->io.axes.y == true
					 && data->io.axes.z == false )
			{
				data->io.axes.x = false; 
				data->io.axes.y = false; 
				data->io.axes.z = true;
				printf ("axes: Z\n");
			}
			else if (data->io.axes.x == false 
					 && data->io.axes.y == false
					 && data->io.axes.z == true )
			{
				data->io.axes.x = true; 
				data->io.axes.y = true; 
				data->io.axes.z = true;
				printf ("axes: XYZ\n");
			}
			break;

		case kNPgraphSelect :
			if (data->map.nodeRootIndex < kNPnodeRootPin)	//zz debug, does not support cam and grid
				return;
			if (node->selected)
				node->selected = false;
			else
				node->selected = true;
			break;
		case kNPgraphSelectOn :
			node->selected = true;
			break;
		case kNPgraphSelectOff :
			node->selected = false;
			break;

		case kNPgraphSelectAll :
			if (data->map.nodeRootIndex < kNPnodeRootPin)		// if cam or grid then ...
				npCtrlCommand (kNPgraphNext, data);		// select our previous node
			if (data->map.selectAll)
			{
				data->map.selectAll = false;
				commandTemp = kNPgraphSelectOff;
			}
			else
			{
				data->map.selectAll = true;
				commandTemp = kNPgraphSelectOn;
			}
			
			data->map.selectedPinNode = data->map.activeNode;		//zz debug, should not be necessary
			
			for (i=kNPnodeRootPin; i < data->map.nodeRootCount; i++)
				TraverseTree (commandTemp, data->map.node[i], data);
			
			data->map.activeNode = data->map.selectedPinNode;
			break;
		case kNPgraphSelectAllOff :
			if (data->map.nodeRootIndex < kNPnodeRootPin)	// if cam or grid then ...
				npCtrlCommand (kNPgraphNext, data);		// select our previous node
	
			data->map.selectAll = false;
			commandTemp = kNPgraphSelectOff;
			
			data->map.selectedPinNode = data->map.activeNode;
			
			for (i=kNPnodeRootPin; i < data->map.nodeRootCount; i++)
				TraverseTree (commandTemp, data->map.node[i], data);
			
			data->map.activeNode = data->map.selectedPinNode;
			break;
		case kNPgraphSelectNone :
			data->map.selectionSet.x = false;
			data->map.selectionSet.y = false;
			data->map.selectionSet.z = false;
			break;
		case kNPgraphSelectOne :
//			data->map.selectionSet.x = 1 - data->map.selectionSet.x;	//zz debug
			if (data->io.key.modShift)
				npFileSaveMap  ("ANTZ0001.CSV", 1, 12, dataRef);
			else
				npFileOpenMap ("ANTZ0001.CSV", 1, 12, dataRef);
			break;
		case kNPgraphSelectTwo :
//			data->map.selectionSet.y = 1 - data->map.selectionSet.y;
			if (data->io.key.modShift)
				npFileSaveMap  ("ANTZ0002.CSV", 1, 12, dataRef);
			else
				npFileOpenMap ("ANTZ0002.CSV", 1, 12, dataRef);
			break;
		case kNPgraphSelectThree :
//			data->map.selectionSet.z = 1 - data->map.selectionSet.z;
			if (data->io.key.modShift)
				npFileSaveMap  ("ANTZ0003.CSV", 1, 12, dataRef);
			else
				npFileOpenMap ("ANTZ0003.CSV", 1, 12, dataRef);			//zz end
			break;

		default : break;
	}
}

//------------------------------------------------------------------------------
npChannel (int command, NPnodePtr node, void* dataRef)
{
	pData data = (pData) dataRef;

//	NPnodePtr node = NULL;
	if (command == kNPgraphChannelDown)
	{
		if (data->io.axes.x)
		{
			node->channel.x--; 
			if (node->channel.x < 0)
				node->channel.x = data->io.channelCount - 1;
		}
		if (data->io.axes.y)
		{
			node->channel.y--; 
			if (node->channel.y < 0)
				node->channel.y = data->io.channelCount - 1;
		}
		if (data->io.axes.z)
		{
			node->channel.z--; 
			if (node->channel.z < 0)
				node->channel.z = data->io.channelCount - 1;
		}
	}
	else if (command == kNPgraphChannelUp)
	{
		if (data->io.axes.x)
		{
			node->channel.x++; 
			if (node->channel.x >= data->io.channelCount)
				node->channel.x = 0;
		}
		if (data->io.axes.y)
		{
			node->channel.y++; 
			if (node->channel.y >= data->io.channelCount)
				node->channel.y = 0;
		}
		if (data->io.axes.z)
		{
			node->channel.z++; 
			if (node->channel.z >= data->io.channelCount)
				node->channel.z = 0;
		}
	}

	printf ("X: %d   Y: %d   Z: %d\n", 
			node->channel.x, node->channel.y, node->channel.z);
}

//------------------------------------------------------------------------------
#define kNPfacetCube 6
void npCtrlProperty (int command, void* dataRef)
{
	pData data = (pData) dataRef;

	NPnodePtr nodeParent = NULL;
	NPnodePtr node = NULL;

//	if (data->map.nodeRootIndex < kNPnodeRootPin)
//		node = data->map.node[data->map.nodeRootIndex];
//	else
	node = data->map.activeNode;

	switch (command)
	{
		case kNPgraphChannelUp : npChannel (command, node, data); break;
		case kNPgraphChannelDown : npChannel (command, node, data); break;

		case kNPgraphColorUp :
			node->colorIndex++;
			SetIndexColor (&node->color, &node->colorIndex);
			printf ("color index: %d   r: %d   g: %d   b: %d\n",
				node->colorIndex, node->color.r, node->color.g, node->color.b);
			break;
		case kNPgraphColorDown :
			node->colorIndex--;
			SetIndexColor (&node->color, &node->colorIndex);
			printf ("color index: %d   r: %d   g: %d   b: %d\n",
				node->colorIndex, node->color.r, node->color.g, node->color.b);
			break;

		case kNPgraphAlphaUp :
			if ((node->color.a + 10) > 255)
				node->color.a = 255;
			else
				node->color.a += 10;
			printf ("alpha: %d\n",node->color.a);
			break;
		case kNPgraphAlphaDown :
			if ((node->color.a - 10) < 0)
				node->color.a = 0;
			else
				node->color.a -= 10;
			printf ("alpha: %d\n",node->color.a);
			break;

		case kNPgraphTexture :
			if (data->io.key.modShift)
				node->textureID--;
			else
				node->textureID++;
			if (node->textureID < 0)
				node->textureID = data->io.gl.textureCount;
			if (node->textureID > data->io.gl.textureCount)
				node->textureID = 0;

			printf ("textureID: %d\n", node->textureID);
			break;

//		case kNPgraphCenter : node->center = 1 - node->center; break;

		case kNPgraphLines :
			node->lineWidth += 1.0f;
			if (node->lineWidth > 5.0f)
				node->lineWidth = 0.0f;
			break;
		case kNPgraphPoints :
			node->pointSize += 1.0f;
			if (node->pointSize > 5.0f)
				node->pointSize = 0;
			break;

		case kNPgraphPrimitive :
			if (data->io.key.modShift)
				node->geometry--;
			else
				node->geometry++;

			if (node->geometry > kNPprimitivePin)
				node->geometry = 0;
			if (node->geometry < 0)
				node->geometry = kNPprimitivePin;
			printf ("geometry: %d\n", node->geometry);
			break;

		case kNPgraphTopo :
			if (data->io.key.modShift)
			{
				node->facet++;
				if (node->facet > kNPfacetCube)
					node->facet = 1;
				printf ("facet: %d\n", node->facet); 
			}
			else
			{	
				node->topo++;

				if (node->topo > 4)//kNPtopoCount)						debug zz
					node->topo = 0;
		
				switch (node->topo)
				{
					case kNPtopoNull : printf	("topo: %d default\n", node->topo); break;
					case kNPtopoCube : 
						node->geometry = kNPglutSolidCube;
						printf ("topo: %d cube\n", node->topo);
						break;
					case kNPtopoSphere :
						node->geometry = kNPglutSolidSphere;
						printf ("topo: %d sphere\n", node->topo); 
						break;
					case kNPtopoTorus : 
						node->geometry = kNPglutSolidTorus;
						printf ("topo: %d torus\n", node->topo); 
						break;
					case kNPtopoSurface : node->topo++; //skip this type for now debug zz //printf("topo: %d surface\n", node->topo); break;
					case kNPtopoPin : 
						node->geometry = kNPprimitivePin;
						printf ("topo: %d pin\n", node->topo);
						break;
					case kNPtopoGrid :
				//		node->geometry = kNPgeoGrid;
						printf ("topo: %d grid\n", node->topo);
						break;
					default : printf ("topo: %d other\n", node->topo); break;
				}
			}
			break;

		case kNPgraphShader :
			if (data->io.key.modShift)
				node->shader--;
			else
				node->shader++;

			// handle roll-over, only allow root node to be a pin
			if (node->branchLevel == 0)	
			{
				if (node->shader > kNPprimitivePin)
					node->shader = 0;
				if (node->shader < 0)
					node->shader = kNPprimitivePin;
			}
			else
			{
				if (node->shader >= kNPprimitivePin)
					node->shader = 0;
				if (node->shader < 0)
					node->shader = kNPprimitivePin - 1;
			}
			break;

//		case kNPgraphMesh : node->geometry = kSurfaceMesh; break;
//		case kNPgraphScroll : node->scroll = 1 - node->scroll; break;
		case kNPgraphFreeze :
			node->freeze = 1 - node->freeze;
			if (node->freeze)
				printf ("freeze\n", node->freeze);
			else
				printf ("unfrozen\n", node->freeze);
			break;
		case kNPgraphHide :
			node->hide = 1 - node->hide;
			if (node->hide)
				printf ("hide\n", node->hide);
			else
				printf ("unhide\n", node->hide);
			break;

		case kNPgraphClear :
			node->triggerOnLo.x = false;
			node->triggerOnLo.y = false;
			node->triggerOnLo.z = false;
			node->triggerOnHi.x = false;
			node->triggerOnHi.y = false;
			node->triggerOnHi.z = false;
			printf ("Clear Limits\n");
			break;

		case kNPgraphSampleInterval : break;

		case kNPgraphSegments : 
			if (data->io.key.modShift)
			{
				if (data->io.axes.x)
				{
					node->segments.x--;
					if (node->segments.x < 0)
						node->segments.x = 0;
				}
				if (data->io.axes.y)
				{
					node->segments.y--;
					if (node->segments.y < 0)
						node->segments.y = 0;
				}
				if (data->io.axes.z)
				{
					node->segments.z--;
					if (node->segments.z < 0)
						node->segments.z = 0;
				}
			}
			else
			{
				if (data->io.axes.x)
					node->segments.x++;
				if (data->io.axes.y)
					node->segments.y++;
				if (data->io.axes.z)
					node->segments.z++;
			}
			printf("segments x: %d   y: %d   z: %d\n",
					node->segments.x, node->segments.y, node->segments.z );
			break;

		case kEventLowSetpoint :
			printf ("Set Lower Limit ");
			if (data->io.key.modShift)
			{
				node->triggerOnLo.x = false;
				node->triggerOnLo.y = false;
				node->triggerOnLo.z = false;
				printf ("OFF\n");
			}
			else
			{
				if (data->io.axes.x)
				{
					node->triggerLo.x = node->translate.x;
					node->triggerOnLo.x = true;
					printf("x: %f   ", node->triggerLo.x);
				}
				if (data->io.axes.y)
				{
					node->triggerLo.y = node->translate.y;
					node->triggerOnLo.y = true;
					printf("y: %f   ", node->triggerLo.y);
				}
				if (data->io.axes.z)
				{
					node->triggerLo.z = node->translate.z;
					node->triggerOnLo.z = true;
					printf("z: %f   ", node->triggerLo.z);
				}
				printf ("\n");   
			}
			break;
		case kEventHighSetpoint :
			printf ("Set Upper Limit ");
			if (data->io.key.modShift)
			{
				node->triggerOnHi.x = false;
				node->triggerOnHi.y = false;
				node->triggerOnHi.z = false;
				printf ("OFF\n");
			}
			else
			{
				if (data->io.axes.x)
				{
					node->triggerHi.x = node->translate.x;
					node->triggerOnHi.x = true;
					printf("x: %f   ", node->triggerHi.x);
				}
				if (data->io.axes.y)
				{
					node->triggerHi.y = node->translate.y;
					node->triggerOnHi.y = true;
					printf("y: %f   ", node->triggerHi.y);
				}
				if (data->io.axes.z)
				{
					node->triggerHi.z = node->translate.z;
					node->triggerOnHi.z = true;
					printf("z: %f   ", node->triggerHi.z);
				}
				printf ("\n");   
			}
			break;

		case kNPgraphDrawIndicators :
		//	node->drawLabel = 1 - node->drawLabel;						//debug, zz
		/*	if (data->io.axes.x)
				node->drawCurrentValue.x;
			if (data->io.axes.y)
				node->drawCurrentValue.y;
			if (data->io.axes.z)
				node->drawCurrentValue.z;
		*/	
			break;

		case kNPgraphRatio :					// torus inner radius 
			if (data->io.key.modShift)
			{
				node->ratio -= 0.01f;
				if (node->ratio < 0.01f)
					node->ratio = 0.01f;
			}
			else
			{
				node->ratio += 0.01f;
				if (node->ratio > 1.0f)
					node->ratio = 1.0f;
			}
			printf ("ratio: %f\n", node->ratio);
			break;

		default : break;
	}
}


//------------------------------------------------------------------------------
void npCtrlGlobal (int command, void* dataRef)
{
	pData data = (pData) dataRef;

	NPnodePtr node = data->map.node[data->map.nodeRootIndex];

	switch (command)
	{
		case kNPgraphMenu : data->io.drawMenu = 1 - data->io.drawMenu; break;

		case kFullScreenExit : npGlutFullScreenExit(); break;

		case kNPgraphClearFrameBuffer :
			data->io.clearFrameBuffer = 1 - data->io.clearFrameBuffer;
			break;

		case kNPgraphBackground :
			if (data->io.clearIndex)
			{
				data->io.clear.r = 0.0f;
				data->io.clear.g = 0.0f;
				data->io.clear.b = 0.0f;
				data->io.clear.a = 0.0f;
				
				data->io.clearIndex = 0;
				printf ("Background Black\n");
			}
			else
			{
				data->io.clear.r = 1.0f;
				data->io.clear.g = 1.0f;
				data->io.clear.b = 1.0f;
				data->io.clear.a = 1.0f;

				data->io.clearIndex = 1;
				printf ("Background White\n");
			}
			break;

		case kNPgraphNormalMode : 
			data->io.gl.normal++;
				if (data->io.gl.normal >= kNPglNormalSize)
				{
					data->io.gl.normal = kNPglNormalNull;
					printf ("Disable Normalize: no rescaling\n");
				}
				else
					printf ("Enable Normalize: rescale normals\n");
			break;

		case kNPgraphShaderMode :
			data->io.gl.shade++;
			if (data->io.gl.shade >= kNPglShadeSize)
				data->io.gl.shade = kNPglShadeSmooth;
			printf ("shader mode: %d\n", data->io.gl.alphaMode);
			break;

		case kNPgraphAlphaMode :
			data->io.gl.alphaMode++;
			if (data->io.gl.alphaMode >= kNPalphaModeCount)
				data->io.gl.alphaMode = kNPalphaModeNull;

			printf ("alpha mode: %d Transparency ", data->io.gl.alphaMode);
			switch (data->io.gl.alphaMode)
			{
				case kNPalphaModeNull : printf("OFF\n"); break;
				case kNPalphaModeSubtractive : printf("Subtractive\n"); break;
				case kNPalphaModeAdditive : printf("Additive\n"); break;
				case kNPalphaModeDark : printf("Dark\n"); break;
			}
			break;
		default : break;
	}
}


//------------------------------------------------------------------------------
void npUpdateCtrl (void* dataRef)
{
	bool flag = false;
	int i = 0;
	static int firstRunLoop = true;

	pData data = (pData) dataRef;

	npUpdateData (dataRef);			//update our data map
	npUpdateIO (dataRef);			//update IO, mouse, sound, txt, video...
	npUpdateEngine (dataRef);		//position, physics, interactions...

	// zero out our mouse to prevent drifting objects
	data->io.mouse.delta.x = 0.0f;											//debug, zz
	data->io.mouse.delta.y = 0.0f;

	// check for command line arguements if first runloop
	if (firstRunLoop)
	{
		firstRunLoop = false;

		// if exist, load command line data files passed in
		for (i=1; i < data->io.argc; i++)		//start at arg 1, 0 is the exe name
		{
			printf("%s\n", data->io.argv[i]);

			// check to see if URL, if so set the flag
			// supports both http and local files
			if ( strncmp("http", data->io.argv[i], 4) == 0 )
	//			|| strncmp("file:///", data->io.argv[i], 8) == 0 ) //debug, zz cannot pass parameters to local file
				flag = true;

			printf ("flag: %d \n", flag);
			// if URL then copy the string, otherwise assume its a CSV file
			if (flag)
				strcpy( data->io.url, data->io.argv[i] );
			else
				npFileOpenMap (data->io.argv[i], 1, 13, dataRef);
		}
	}
}

